import sys
import csv
import time
import serial
import struct
import collections
import threading

PACKET_DEF = {'i1': [116, bytes([0x55, 0x55, 0x69, 0x31])]} # {'<packet name>': [<payload length>, <packet header>]}

class OpenRTK:
    def __init__(self, port, baud=460800):
        self.port = port
        self.baud = baud
        self.ser = serial.Serial(self.port, self.baud, timeout=0.5)
        self.myqueue = collections.deque(maxlen=10000)

    def start_listen(self):
        '''
        Capture the packet from the buffer and add it to the queue
        '''
        header = PACKET_DEF['i1'][1]
        data_len = PACKET_DEF['i1'][0]
        packet_len = 4 + 1 + data_len + 2
        packet_data_part = None
        while True:
            data = self.ser.read(packet_len)
            if packet_data_part is not None:
                part_len = len(packet_data_part)
                packet_data_part_ex = data[: packet_len-part_len]
                packet_data = packet_data_part + packet_data_part_ex
                packet_data_part = None
                if len(packet_data) < 123:
                    print(packet_data)
                self.myqueue.append(packet_data)
            for i in range(packet_len):
                if data[i: i+4] == header and i + packet_len <= packet_len:
                    packet_data = data[i: i+packet_len]
                    packet_data_part = None
                    self.myqueue.append(packet_data)
                elif data[i: i+4] == header and i + packet_len > packet_len:
                    packet_data_part = data[i:]

    def parse_data(self):
        '''
        Parse the packet data in queue and log it to csv file
        '''
        timestamp = time.strftime('%Y_%m_%d_%H_%M_%S')
        logf_name = f'il_packet_parse_data_{timestamp}.csv'
        print('Data log start')
        self.data_logger(logf_name, is_init=True)
        cnt = 0
        while True:
            if len(self.myqueue) != 0:
                packet_data = self.myqueue.popleft()
                crc = packet_data[-2:]
                cal_crc = bytes(self.calc_crc(packet_data[2: -2]))
                if cal_crc != crc:
                    print('CRC ERROR!')
                    sys.exit()
                payload = packet_data[5:-2]
                latest = self.il_parser(payload)
                record = [latest[0], latest[1], latest[2], latest[3], latest[4], latest[5], latest[6], latest[7], latest[8], latest[9], latest[10], latest[11], latest[12], latest[13]\
                    , latest[14], latest[15], latest[16], latest[17], latest[18], latest[19], latest[20], latest[21]]
                self.data_logger(logf_name, record)
                cnt += 1
                if cnt <= 50:
                    print(f'>>', end="")
                else:
                    print("\033[2K\r", end="")
                    cnt = 0


    def flush_buffer_data(self):
        '''
        use to flush the buffer
        '''
        self.UUT.flushOutput()
        self.UUT.flushInput()
        time.sleep(0.01)

    def il_parser(self, payload):
        '''
        packet parse function
        '''
        fmt = '<HIBBdddddddddfffffffff'
        data = struct.unpack(fmt, payload)
        gps_week = data[0]
        gps_millisecs = data[1] / 1000
        ins_status = data[2]
        ins_position_type = data[3]
        latitude = data[4]
        longitude = data[5]
        height = data[6]
        north_velocity = data[7]
        east_velocity = data[8]
        up_velocity = data[9]
        roll = data[10]
        pitch = data[11]
        heading = data[12]
        latitude_std = data[13]
        longitude_std = data[14]
        height_std = data[15]
        north_velocity_std = data[16]
        east_velocity_std = data[17]
        up_velocity_std = data[18]
        roll_std = data[19]
        pitch_std = data[20]
        heading_std = data[21]

        return gps_week, gps_millisecs, ins_status, ins_position_type, latitude, longitude, height, north_velocity_std, east_velocity_std, \
            up_velocity_std, roll, pitch, heading, latitude_std, longitude_std, height_std, north_velocity_std, east_velocity_std, \
            up_velocity_std, roll_std, pitch_std, heading_std 

    def data_logger(self, logf_name, data=None, is_init=False):
        if is_init:
            with open(logf_name, 'a', newline='') as csvfile:
                writer = csv.writer(csvfile, delimiter=',')
                header = ['gps_week', 'gps_millisecs', 'ins_status', 'ins_position_type', 'latitude', 'longitude', 'height', 'north_velocity_std', 'east_velocity_std', 'up_velocity_std', 'roll',\
                    'pitch', 'heading', 'latitude_std', 'longitude_std', 'height_std', 'north_velocity_std', 'east_velocity_std', 'up_velocity_std', 'roll_std', 'pitch_std', 'heading_std']
                writer.writerow(header)
        else:
            with open(logf_name, 'a', newline='') as csvfile:
                writer = csv.writer(csvfile, delimiter=',')
                writer.writerow(data)
    
    def calc_crc(self, payload):
        '''
        Calculates 16-bit CRC-CCITT
        '''
        crc = 0x1D0F
        for bytedata in payload:
            crc = crc ^ (bytedata << 8)
            i = 0
            while i < 8:
                if crc & 0x8000:
                    crc = (crc << 1) ^ 0x1021
                else:
                    crc = crc << 1
                i += 1

        crc = crc & 0xffff
        crc_msb = (crc & 0xFF00) >> 8
        crc_lsb = (crc & 0x00FF)
        return [crc_msb, crc_lsb]

if __name__ == "__main__":
    port = 'COM17' # edit user's port name here
    openrtk = OpenRTK(port=port)
    t_lst = []
    t1 = threading.Thread(target=openrtk.start_listen, name='Data Listener')
    t2 = threading.Thread(target=openrtk.parse_data, name='Data Paser')
    t_lst.append(t1)
    t_lst.append(t2)
    for t in t_lst:
        t.start()
    for t in t_lst:
        t.join()